#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging

import six

from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


@six.add_metaclass(abc.ABCMeta)
class SenderBase(object):
    """Base request/ack/reply sending interface."""

    def __init__(self, conf):
        self.conf = conf

    @abc.abstractmethod
    def send(self, socket, message):
        pass


class RequestSender(SenderBase):
    pass


class ReplySender(SenderBase):
    pass


class RequestSenderProxy(RequestSender):

    def send(self, socket, request):
        socket.send(b'', zmq.SNDMORE)
        socket.send(six.b(str(request.msg_type)), zmq.SNDMORE)
        socket.send(six.b(request.routing_key), zmq.SNDMORE)
        socket.send(six.b(request.message_id), zmq.SNDMORE)
        socket.send_dumped(request.context, zmq.SNDMORE)
        socket.send_dumped(request.message)

        LOG.debug("->[proxy:%(addr)s] Sending %(msg_type)s message "
                  "%(msg_id)s to target %(target)s",
                  {"addr": list(socket.connections),
                   "msg_type": zmq_names.message_type_str(request.msg_type),
                   "msg_id": request.message_id,
                   "target": request.target})


class ReplySenderProxy(ReplySender):

    def send(self, socket, reply):
        LOG.debug("Replying to %s", reply.message_id)

        assert reply.msg_type == zmq_names.REPLY_TYPE, "Reply expected!"

        socket.send(b'', zmq.SNDMORE)
        socket.send(six.b(str(reply.msg_type)), zmq.SNDMORE)
        socket.send(reply.reply_id, zmq.SNDMORE)
        socket.send(reply.message_id, zmq.SNDMORE)
        socket.send_dumped(reply.to_dict())


class RequestSenderDirect(RequestSender):

    def send(self, socket, request):
        socket.send(b'', zmq.SNDMORE)
        socket.send(six.b(str(request.msg_type)), zmq.SNDMORE)
        socket.send_string(request.message_id, zmq.SNDMORE)
        socket.send_dumped(request.context, zmq.SNDMORE)
        socket.send_dumped(request.message)

        LOG.debug("Sending %(msg_type)s message %(msg_id)s to "
                  "target %(target)s",
                  {"msg_type": zmq_names.message_type_str(request.msg_type),
                   "msg_id": request.message_id,
                   "target": request.target})


class ReplySenderDirect(ReplySender):

    def send(self, socket, reply):
        LOG.debug("Replying to %s", reply.message_id)

        assert reply.msg_type == zmq_names.REPLY_TYPE, "Reply expected!"

        socket.send(reply.reply_id, zmq.SNDMORE)
        socket.send(b'', zmq.SNDMORE)
        socket.send_dumped(reply.to_dict())



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names

zmq = zmq_async.import_zmq()


class RoutingTable(object):
    """This class implements local routing-table cache
        taken from matchmaker. Its purpose is to give the next routable
        host id (remote DEALER's id) by request for specific target in
        round-robin fashion.
    """

    def __init__(self, conf, matchmaker):
        self.conf = conf
        self.matchmaker = matchmaker
        self.routing_table = {}
        self.routable_hosts = {}

    def get_all_hosts(self, target):
        self._update_routing_table(target)
        return list(self.routable_hosts.get(str(target)) or [])

    def get_routable_host(self, target):
        self._update_routing_table(target)
        hosts_for_target = self.routable_hosts[str(target)]
        host = hosts_for_target.pop(0)
        if not hosts_for_target:
            self._renew_routable_hosts(target)
        return host

    def _is_tm_expired(self, tm):
        return 0 <= self.conf.zmq_target_expire <= time.time() - tm

    def _update_routing_table(self, target):
        routing_record = self.routing_table.get(str(target))
        if routing_record is None:
            self._fetch_hosts(target)
            self._renew_routable_hosts(target)
        elif self._is_tm_expired(routing_record[1]):
            self._fetch_hosts(target)

    def _fetch_hosts(self, target):
        self.routing_table[str(target)] = (self.matchmaker.get_hosts(
            target, zmq_names.socket_type_str(zmq.DEALER)), time.time())

    def _renew_routable_hosts(self, target):
        hosts, _ = self.routing_table[str(target)]
        self.routable_hosts[str(target)] = list(hosts)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver.client import zmq_senders
from oslo_messaging._drivers.zmq_driver.client import zmq_sockets_manager
from oslo_messaging._drivers.zmq_driver.server.consumers \
    import zmq_consumer_base
from oslo_messaging._drivers.zmq_driver.server import zmq_incoming_message
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_updater
from oslo_messaging._i18n import _LE, _LI

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class DealerConsumer(zmq_consumer_base.SingleSocketConsumer):

    def __init__(self, conf, poller, server):
        self.sender = zmq_senders.ReplySenderProxy(conf)
        self.sockets_manager = zmq_sockets_manager.SocketsManager(
            conf, server.matchmaker, zmq.ROUTER, zmq.DEALER)
        self.host = None
        super(DealerConsumer, self).__init__(conf, poller, server, zmq.DEALER)
        self.connection_updater = ConsumerConnectionUpdater(
            conf, self.matchmaker, self.socket)
        LOG.info(_LI("[%s] Run DEALER consumer"), self.host)

    def subscribe_socket(self, socket_type):
        try:
            socket = self.sockets_manager.get_socket_to_routers()
            self.sockets.append(socket)
            self.host = socket.handle.identity
            self.poller.register(socket, self.receive_message)
            return socket
        except zmq.ZMQError as e:
            LOG.error(_LE("Failed connecting to ROUTER socket %(e)s") % e)
            raise rpc_common.RPCException(str(e))

    def receive_message(self, socket):
        try:
            empty = socket.recv()
            assert empty == b'', 'Bad format: empty delimiter expected'
            reply_id = socket.recv()
            message_type = int(socket.recv())
            message_id = socket.recv()
            context = socket.recv_loaded()
            message = socket.recv_loaded()
            LOG.debug("[%(host)s] Received %(msg_type)s message %(msg_id)s",
                      {"host": self.host,
                       "msg_type": zmq_names.message_type_str(message_type),
                       "msg_id": message_id})
            if message_type == zmq_names.CALL_TYPE:
                return zmq_incoming_message.ZmqIncomingMessage(
                    context, message, reply_id, message_id, socket, self.sender
                )
            elif message_type in zmq_names.NON_BLOCKING_TYPES:
                return zmq_incoming_message.ZmqIncomingMessage(context,
                                                               message)
            else:
                LOG.error(_LE("Unknown message type: %s"),
                          zmq_names.message_type_str(message_type))
        except (zmq.ZMQError, AssertionError) as e:
            LOG.error(_LE("Receiving message failure: %s"), str(e))

    def cleanup(self):
        LOG.info(_LI("[%s] Destroy DEALER consumer"), self.host)
        self.connection_updater.cleanup()
        super(DealerConsumer, self).cleanup()


class ConsumerConnectionUpdater(zmq_updater.ConnectionUpdater):

    def _update_connection(self):
        routers = self.matchmaker.get_routers()
        for router_address in routers:
            self.socket.connect_to_host(router_address)



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import retrying

from oslo_messaging._drivers.zmq_driver.client.publishers.dealer \
    import zmq_dealer_publisher_base
from oslo_messaging._drivers.zmq_driver.client import zmq_receivers
from oslo_messaging._drivers.zmq_driver.client import zmq_senders
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class DealerPublisherDirect(zmq_dealer_publisher_base.DealerPublisherBase):
    """DEALER-publisher using direct connections."""

    def __init__(self, conf, matchmaker):
        sender = zmq_senders.RequestSenderDirect(conf)
        receiver = zmq_receivers.ReplyReceiverDirect(conf)
        super(DealerPublisherDirect, self).__init__(conf, matchmaker, sender,
                                                    receiver)

    def _connect_socket(self, request):
        return self.sockets_manager.get_socket(request.target)

    def _send_non_blocking(self, request):
        try:
            socket = self._connect_socket(request)
        except retrying.RetryError:
            return

        if request.msg_type in zmq_names.MULTISEND_TYPES:
            for _ in range(socket.connections_count()):
                self.sender.send(socket, request)
        else:
            self.sender.send(socket, request)



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import retrying

from oslo_messaging._drivers.zmq_driver.client.publishers.dealer \
    import zmq_dealer_publisher_base
from oslo_messaging._drivers.zmq_driver.client import zmq_receivers
from oslo_messaging._drivers.zmq_driver.client import zmq_routing_table
from oslo_messaging._drivers.zmq_driver.client import zmq_senders
from oslo_messaging._drivers.zmq_driver import zmq_address
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_updater

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class DealerPublisherProxy(zmq_dealer_publisher_base.DealerPublisherBase):
    """DEALER-publisher via proxy."""

    def __init__(self, conf, matchmaker):
        sender = zmq_senders.RequestSenderProxy(conf)
        receiver = zmq_receivers.ReplyReceiverProxy(conf)
        super(DealerPublisherProxy, self).__init__(conf, matchmaker, sender,
                                                   receiver)
        self.socket = self.sockets_manager.get_socket_to_publishers()
        self.routing_table = zmq_routing_table.RoutingTable(self.conf,
                                                            self.matchmaker)
        self.connection_updater = \
            PublisherConnectionUpdater(self.conf, self.matchmaker, self.socket)

    def _connect_socket(self, request):
        return self.socket

    def send_call(self, request):
        try:
            request.routing_key = \
                self.routing_table.get_routable_host(request.target)
        except retrying.RetryError:
            self._raise_timeout(request)
        return super(DealerPublisherProxy, self).send_call(request)

    def _get_routing_keys(self, request):
        try:
            if request.msg_type in zmq_names.DIRECT_TYPES:
                return [self.routing_table.get_routable_host(request.target)]
            else:
                return \
                    [zmq_address.target_to_subscribe_filter(request.target)] \
                    if self.conf.use_pub_sub else \
                    self.routing_table.get_all_hosts(request.target)
        except retrying.RetryError:
            return []

    def _send_non_blocking(self, request):
        for routing_key in self._get_routing_keys(request):
            request.routing_key = routing_key
            self.sender.send(self.socket, request)

    def cleanup(self):
        super(DealerPublisherProxy, self).cleanup()
        self.connection_updater.stop()
        self.socket.close()


class PublisherConnectionUpdater(zmq_updater.ConnectionUpdater):

    def _update_connection(self):
        publishers = self.matchmaker.get_publishers()
        for pub_address, router_address in publishers:
            self.socket.connect_to_host(router_address)



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_socket

zmq = zmq_async.import_zmq()


class SocketsManager(object):

    def __init__(self, conf, matchmaker, listener_type, socket_type):
        self.conf = conf
        self.matchmaker = matchmaker
        self.listener_type = listener_type
        self.socket_type = socket_type
        self.zmq_context = zmq.Context()
        self.outbound_sockets = {}
        self.socket_to_publishers = None
        self.socket_to_routers = None

    def get_hosts(self, target):
        return self.matchmaker.get_hosts(
            target, zmq_names.socket_type_str(self.listener_type))

    @staticmethod
    def _key_from_target(target):
        return target.topic if target.fanout else str(target)

    def _get_hosts_and_connect(self, socket, target):
        hosts = self.get_hosts(target)
        self._connect_to_hosts(socket, target, hosts)

    def _track_socket(self, socket, target):
        key = self._key_from_target(target)
        self.outbound_sockets[key] = (socket, time.time())

    def _connect_to_hosts(self, socket, target, hosts):
        for host in hosts:
            socket.connect_to_host(host)
        self._track_socket(socket, target)

    def _check_for_new_hosts(self, target):
        key = self._key_from_target(target)
        socket, tm = self.outbound_sockets[key]
        if 0 <= self.conf.zmq_target_expire <= time.time() - tm:
            self._get_hosts_and_connect(socket, target)
        return socket

    def get_socket(self, target):
        key = self._key_from_target(target)
        if key in self.outbound_sockets:
            socket = self._check_for_new_hosts(target)
        else:
            socket = zmq_socket.ZmqSocket(self.conf, self.zmq_context,
                                          self.socket_type)
            self._get_hosts_and_connect(socket, target)
        return socket

    def get_socket_to_publishers(self):
        if self.socket_to_publishers is not None:
            return self.socket_to_publishers
        self.socket_to_publishers = zmq_socket.ZmqSocket(
            self.conf, self.zmq_context, self.socket_type)
        publishers = self.matchmaker.get_publishers()
        for pub_address, router_address in publishers:
            self.socket_to_publishers.connect_to_host(router_address)
        return self.socket_to_publishers

    def get_socket_to_routers(self):
        if self.socket_to_routers is not None:
            return self.socket_to_routers
        self.socket_to_routers = zmq_socket.ZmqSocket(
            self.conf, self.zmq_context, self.socket_type)
        routers = self.matchmaker.get_routers()
        for router_address in routers:
            self.socket_to_routers.connect_to_host(router_address)
        return self.socket_to_routers

    def cleanup(self):
        for socket, tm in self.outbound_sockets.values():
            socket.close()



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
from concurrent import futures
import logging

import retrying

import oslo_messaging
from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver.client.publishers \
    import zmq_publisher_base
from oslo_messaging._drivers.zmq_driver.client import zmq_sockets_manager
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._i18n import _LE

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class DealerPublisherBase(zmq_publisher_base.PublisherBase):
    """Abstract DEALER-publisher."""

    def __init__(self, conf, matchmaker, sender, receiver):
        sockets_manager = zmq_sockets_manager.SocketsManager(
            conf, matchmaker, zmq.ROUTER, zmq.DEALER
        )
        super(DealerPublisherBase, self).__init__(sockets_manager, sender,
                                                  receiver)

    @staticmethod
    def _check_pattern(request, supported_pattern):
        if request.msg_type != supported_pattern:
            raise zmq_publisher_base.UnsupportedSendPattern(
                zmq_names.message_type_str(request.msg_type)
            )

    @staticmethod
    def _raise_timeout(request):
        raise oslo_messaging.MessagingTimeout(
            "Timeout %(tout)s seconds was reached for message %(msg_id)s" %
            {"tout": request.timeout, "msg_id": request.message_id}
        )

    @abc.abstractmethod
    def _connect_socket(self, request):
        pass

    def _recv_reply(self, request):
        reply_future, = self.receiver.track_request(request)

        try:
            _, reply = reply_future.result(timeout=request.timeout)
        except AssertionError:
            LOG.error(_LE("Message format error in reply for %s"),
                      request.message_id)
            return None
        except futures.TimeoutError:
            self._raise_timeout(request)
        finally:
            self.receiver.untrack_request(request)

        if reply.failure:
            raise rpc_common.deserialize_remote_exception(
                reply.failure, request.allowed_remote_exmods
            )
        else:
            return reply.reply_body

    def send_call(self, request):
        self._check_pattern(request, zmq_names.CALL_TYPE)

        try:
            socket = self._connect_socket(request)
        except retrying.RetryError:
            self._raise_timeout(request)

        self.sender.send(socket, request)
        self.receiver.register_socket(socket)
        return self._recv_reply(request)

    @abc.abstractmethod
    def _send_non_blocking(self, request):
        pass

    def send_cast(self, request):
        self._check_pattern(request, zmq_names.CAST_TYPE)
        self._send_non_blocking(request)

    def send_fanout(self, request):
        self._check_pattern(request, zmq_names.CAST_FANOUT_TYPE)
        self._send_non_blocking(request)

    def send_notify(self, request):
        self._check_pattern(request, zmq_names.NOTIFY_TYPE)
        self._send_non_blocking(request)



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging

import six

from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._i18n import _LE

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class UnsupportedSendPattern(rpc_common.RPCException):

    """Exception to raise from publishers in case of unsupported
    sending pattern called.
    """

    def __init__(self, pattern_name):
        """Construct exception object

        :param pattern_name: Message type name from zmq_names
        :type pattern_name: str
        """
        errmsg = _LE("Sending pattern %s is unsupported.") % pattern_name
        super(UnsupportedSendPattern, self).__init__(errmsg)


@six.add_metaclass(abc.ABCMeta)
class PublisherBase(object):

    """Abstract publisher class

    Each publisher from zmq-driver client should implement
    this interface to serve as a messages publisher.

    Publisher can send request objects from zmq_request.
    """

    def __init__(self, sockets_manager, sender, receiver):

        """Construct publisher

        Accept sockets manager, sender and receiver objects.

        :param sockets_manager: sockets manager object
        :type sockets_manager: zmq_sockets_manager.SocketsManager
        :param senders: request sender object
        :type senders: zmq_senders.RequestSender
        :param receiver: reply receiver object
        :type receiver: zmq_receivers.ReplyReceiver
        """
        self.sockets_manager = sockets_manager
        self.conf = sockets_manager.conf
        self.matchmaker = sockets_manager.matchmaker
        self.sender = sender
        self.receiver = receiver

    @abc.abstractmethod
    def send_call(self, request):
        pass

    @abc.abstractmethod
    def send_cast(self, request):
        pass

    @abc.abstractmethod
    def send_fanout(self, request):
        pass

    @abc.abstractmethod
    def send_notify(self, request):
        pass

    def cleanup(self):
        """Cleanup publisher. Close allocated connections."""
        self.receiver.stop()
        self.sockets_manager.cleanup()



#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging
import threading

import futurist
import six

from oslo_messaging._drivers.zmq_driver.client import zmq_response
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


@six.add_metaclass(abc.ABCMeta)
class ReceiverBase(object):
    """Base response receiving interface."""

    def __init__(self, conf):
        self.conf = conf
        self._lock = threading.Lock()
        self._requests = {}
        self._poller = zmq_async.get_poller()
        self._executor = zmq_async.get_executor(method=self._run_loop)
        self._executor.execute()

    @abc.abstractproperty
    def message_types(self):
        """A list of supported incoming response types."""

    def register_socket(self, socket):
        """Register a socket for receiving data."""
        self._poller.register(socket, recv_method=self.recv_response)

    @abc.abstractmethod
    def recv_response(self, socket):
        """Receive a response and return a tuple of the form
        (reply_id, message_type, message_id, response).
        """

    def track_request(self, request):
        """Track a request via already registered sockets and return
        a list of futures for monitoring all types of responses.
        """
        futures = []
        for message_type in self.message_types:
            future = futurist.Future()
            self._set_future(request.message_id, message_type, future)
            futures.append(future)
        return futures

    def untrack_request(self, request):
        """Untrack a request and stop monitoring any responses."""
        for message_type in self.message_types:
            self._pop_future(request.message_id, message_type)

    def stop(self):
        self._poller.close()
        self._executor.stop()

    def _get_future(self, message_id, message_type):
        with self._lock:
            return self._requests.get((message_id, message_type))

    def _set_future(self, message_id, message_type, future):
        with self._lock:
            self._requests[(message_id, message_type)] = future

    def _pop_future(self, message_id, message_type):
        with self._lock:
            return self._requests.pop((message_id, message_type), None)

    def _run_loop(self):
        data, socket = self._poller.poll(timeout=self.conf.rpc_poll_timeout)
        if data is None:
            return
        reply_id, message_type, message_id, response = data
        assert message_type in self.message_types, \
            "%s is not supported!" % zmq_names.message_type_str(message_type)
        future = self._get_future(message_id, message_type)
        if future is not None:
            LOG.debug("Received %(msg_type)s for %(msg_id)s",
                      {"msg_type": zmq_names.message_type_str(message_type),
                       "msg_id": message_id})
            future.set_result((reply_id, response))


class AckReceiver(ReceiverBase):

    message_types = (zmq_names.ACK_TYPE,)


class ReplyReceiver(ReceiverBase):

    message_types = (zmq_names.REPLY_TYPE,)


class ReplyReceiverProxy(ReplyReceiver):

    def recv_response(self, socket):
        empty = socket.recv()
        assert empty == b'', "Empty expected!"
        reply_id = socket.recv()
        assert reply_id is not None, "Reply ID expected!"
        message_type = int(socket.recv())
        assert message_type == zmq_names.REPLY_TYPE, "Reply expected!"
        message_id = socket.recv()
        raw_reply = socket.recv_loaded()
        assert isinstance(raw_reply, dict), "Dict expected!"
        reply = zmq_response.Response(**raw_reply)
        LOG.debug("Received reply for %s", message_id)
        return reply_id, message_type, message_id, reply


class ReplyReceiverDirect(ReplyReceiver):

    def recv_response(self, socket):
        empty = socket.recv()
        assert empty == b'', "Empty expected!"
        raw_reply = socket.recv_loaded()
        assert isinstance(raw_reply, dict), "Dict expected!"
        reply = zmq_response.Response(**raw_reply)
        LOG.debug("Received reply for %s", reply.message_id)
        return reply.reply_id, reply.msg_type, reply.message_id, reply


class AckAndReplyReceiver(ReceiverBase):

    message_types = (zmq_names.ACK_TYPE, zmq_names.REPLY_TYPE)



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import six

from oslo_messaging._drivers.zmq_driver.proxy import zmq_publisher_proxy
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._drivers.zmq_driver import zmq_socket
from oslo_messaging._drivers.zmq_driver import zmq_updater
from oslo_messaging._i18n import _LI

zmq = zmq_async.import_zmq()
LOG = logging.getLogger(__name__)


class UniversalQueueProxy(object):

    def __init__(self, conf, context, matchmaker):
        self.conf = conf
        self.context = context
        super(UniversalQueueProxy, self).__init__()
        self.matchmaker = matchmaker
        self.poller = zmq_async.get_poller()

        port = conf.zmq_proxy_opts.frontend_port
        host = conf.zmq_proxy_opts.host
        self.fe_router_socket = zmq_socket.ZmqFixedPortSocket(
            conf, context, zmq.ROUTER, host,
            conf.zmq_proxy_opts.frontend_port) if port != 0 else \
            zmq_socket.ZmqRandomPortSocket(conf, context, zmq.ROUTER, host)

        port = conf.zmq_proxy_opts.backend_port
        self.be_router_socket = zmq_socket.ZmqFixedPortSocket(
            conf, context, zmq.ROUTER, host,
            conf.zmq_proxy_opts.backend_port) if port != 0 else \
            zmq_socket.ZmqRandomPortSocket(conf, context, zmq.ROUTER, host)

        self.poller.register(self.fe_router_socket.handle,
                             self._receive_in_request)
        self.poller.register(self.be_router_socket.handle,
                             self._receive_in_request)

        self.pub_publisher = zmq_publisher_proxy.PublisherProxy(
            conf, matchmaker)

        self._router_updater = RouterUpdater(
            conf, matchmaker, self.pub_publisher.host,
            self.fe_router_socket.connect_address,
            self.be_router_socket.connect_address)

    def run(self):
        message, socket = self.poller.poll()
        if message is None:
            return

        msg_type = message[0]
        if self.conf.use_pub_sub and msg_type in (zmq_names.CAST_FANOUT_TYPE,
                                                  zmq_names.NOTIFY_TYPE):
            self.pub_publisher.send_request(message)
        else:
            self._redirect_message(self.be_router_socket.handle
                                   if socket is self.fe_router_socket.handle
                                   else self.fe_router_socket.handle, message)

    @staticmethod
    def _receive_in_request(socket):
        try:
            reply_id = socket.recv()
            assert reply_id is not None, "Valid id expected"
            empty = socket.recv()
            assert empty == b'', "Empty delimiter expected"
            msg_type = int(socket.recv())
            routing_key = socket.recv()
            payload = socket.recv_multipart()
            payload.insert(0, reply_id)
            payload.insert(0, routing_key)
            payload.insert(0, msg_type)
            return payload
        except (AssertionError, zmq.ZMQError):
            LOG.error("Received message with wrong format")
            return None

    @staticmethod
    def _redirect_message(socket, multipart_message):
        message_type = multipart_message.pop(0)
        routing_key = multipart_message.pop(0)
        reply_id = multipart_message.pop(0)
        message_id = multipart_message[0]
        socket.send(routing_key, zmq.SNDMORE)
        socket.send(b'', zmq.SNDMORE)
        socket.send(reply_id, zmq.SNDMORE)
        socket.send(six.b(str(message_type)), zmq.SNDMORE)
        LOG.debug("Dispatching %(msg_type)s message %(msg_id)s - to %(rkey)s" %
                  {"msg_type": zmq_names.message_type_str(message_type),
                   "msg_id": message_id,
                   "rkey": routing_key})
        socket.send_multipart(multipart_message)

    def cleanup(self):
        self.fe_router_socket.close()
        self.be_router_socket.close()
        self.pub_publisher.cleanup()
        self._router_updater.cleanup()


class RouterUpdater(zmq_updater.UpdaterBase):
    """This entity performs periodic async updates
    from router proxy to the matchmaker.
    """

    def __init__(self, conf, matchmaker, publisher_address, fe_router_address,
                 be_router_address):
        self.publisher_address = publisher_address
        self.fe_router_address = fe_router_address
        self.be_router_address = be_router_address
        super(RouterUpdater, self).__init__(conf, matchmaker,
                                            self._update_records)

    def _update_records(self):
        self.matchmaker.register_publisher(
            (self.publisher_address, self.fe_router_address),
            expire=self.conf.zmq_target_expire)
        LOG.info(_LI("[PUB:%(pub)s, ROUTER:%(router)s] Update PUB publisher"),
                 {"pub": self.publisher_address,
                  "router": self.fe_router_address})
        self.matchmaker.register_router(self.be_router_address,
                                        expire=self.conf.zmq_target_expire)
        LOG.info(_LI("[Backend ROUTER:%(router)s] Update ROUTER"),
                 {"router": self.be_router_address})

    def cleanup(self):
        super(RouterUpdater, self).cleanup()
        self.matchmaker.unregister_publisher(
            (self.publisher_address, self.fe_router_address))
        self.matchmaker.unregister_router(
            self.be_router_address)



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from oslo_messaging._drivers.zmq_driver.client import zmq_senders
from oslo_messaging._drivers.zmq_driver.server.consumers \
    import zmq_consumer_base
from oslo_messaging._drivers.zmq_driver.server import zmq_incoming_message
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names
from oslo_messaging._i18n import _LE, _LI

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class RouterConsumer(zmq_consumer_base.SingleSocketConsumer):

    def __init__(self, conf, poller, server):
        self.sender = zmq_senders.ReplySenderDirect(conf)
        super(RouterConsumer, self).__init__(conf, poller, server, zmq.ROUTER)
        LOG.info(_LI("[%s] Run ROUTER consumer"), self.host)

    def _receive_request(self, socket):
        reply_id = socket.recv()
        empty = socket.recv()
        assert empty == b'', 'Bad format: empty delimiter expected'
        msg_type = int(socket.recv())
        message_id = socket.recv_string()
        context = socket.recv_loaded()
        message = socket.recv_loaded()
        return reply_id, msg_type, message_id, context, message

    def receive_message(self, socket):
        try:
            reply_id, msg_type, message_id, context, message = \
                self._receive_request(socket)
            LOG.debug("[%(host)s] Received %(msg_type)s message %(msg_id)s",
                      {"host": self.host,
                       "msg_type": zmq_names.message_type_str(msg_type),
                       "msg_id": message_id})

            if msg_type == zmq_names.CALL_TYPE:
                return zmq_incoming_message.ZmqIncomingMessage(
                    context, message, reply_id, message_id, socket, self.sender
                )
            elif msg_type in zmq_names.NON_BLOCKING_TYPES:
                return zmq_incoming_message.ZmqIncomingMessage(context,
                                                               message)
            else:
                LOG.error(_LE("Unknown message type: %s"),
                          zmq_names.message_type_str(msg_type))
        except (zmq.ZMQError, AssertionError) as e:
            LOG.error(_LE("Receiving message failed: %s"), str(e))

    def cleanup(self):
        LOG.info(_LI("[%s] Destroy ROUTER consumer"), self.host)
        super(RouterConsumer, self).cleanup()



#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import logging

from oslo_messaging._drivers import base
from oslo_messaging._drivers import common as rpc_common
from oslo_messaging._drivers.zmq_driver.client import zmq_response
from oslo_messaging._drivers.zmq_driver import zmq_async
from oslo_messaging._drivers.zmq_driver import zmq_names

LOG = logging.getLogger(__name__)

zmq = zmq_async.import_zmq()


class ZmqIncomingMessage(base.RpcIncomingMessage):

    def __init__(self, context, message, reply_id=None, message_id=None,
                 socket=None, sender=None):

        if sender is not None:
            assert socket is not None, "Valid socket expected!"
            assert message_id is not None, "Valid message ID expected!"
            assert reply_id is not None, "Valid reply ID expected!"

        super(ZmqIncomingMessage, self).__init__(context, message)

        self.reply_id = reply_id
        self.message_id = message_id
        self.socket = socket
        self.sender = sender

    def acknowledge(self):
        """Not sending acknowledge"""

    def reply(self, reply=None, failure=None):
        if self.sender is not None:
            if failure is not None:
                failure = rpc_common.serialize_remote_exception(failure)
            reply = zmq_response.Response(msg_type=zmq_names.REPLY_TYPE,
                                          message_id=self.message_id,
                                          reply_id=self.reply_id,
                                          reply_body=reply,
                                          failure=failure)
            self.sender.send(self.socket, reply)

    def requeue(self):
        """Requeue is not supported"""
